
#===== 3DS Games =====#

class Friends3DS:
	TITLE_ID_EUR = 0x0004013000003202
	TITLE_ID_USA = 0x0004013000003202
	TITLE_ID_JAP = 0x0004013000003202
	LATEST_VERSION = 20

	# Friends 3DS has no product code
	PRODUCT_CODE_EUR = "----"
	PRODUCT_CODE_USA = "----"
	PRODUCT_CODE_JAP = "----"
	
	GAME_SERVER_ID = 0x3200
	ACCESS_KEY = "ridfebb9"
	NEX_VERSION = 20000


#===== Wii U Games =====#

class Friends:
	TITLE_ID_EUR = 0x10001C00
	TITLE_ID_USA = 0x10001C00
	TITLE_ID_JAP = 0x10001C00
	LATEST_VERSION = 0
	
	GAME_SERVER_ID = 0x3200
	ACCESS_KEY = "ridfebb9"
	NEX_VERSION = 20000

class DKCTF:
	TITLE_ID_EUR = 0x0005000010138300
	TITLE_ID_USA = 0x0005000010137F00
	TITLE_ID_JAP = 0x0005000010144800
	LATEST_VERSION = 17
	
	GAME_SERVER_ID = 0x10144800
	ACCESS_KEY = "7fcf384a"
	NEX_VERSION = 30400 #3.4.0
	
class Minecraft:
	TITLE_ID_EUR = 0x00050000101D7500
	TITLE_ID_USA = 0x00050000101D9D00
	TITLE_ID_JAP = 0x00050000101DBE00
	LATEST_VERSION = 688
	
	GAME_SERVER_ID = 0x101D9D00
	ACCESS_KEY = "f1b61c8e"
	NEX_VERSION = 30901 #3.9.1
	
class MK8:
	TITLE_ID_EUR = 0x000500001010ED00
	TITLE_ID_USA = 0x000500001010EC00
	TITLE_ID_JAP = 0x000500001010EB00
	LATEST_VERSION = 64
	
	GAME_SERVER_ID = 0x1010EB00
	ACCESS_KEY = "25dbf96a"
	NEX_VERSION = 30504 #3.5.4 (AMK patch)
	
class PokkenTournament:
	TITLE_ID_EUR = 0x00050000101DF400
	TITLE_ID_USA = 0x00050000101DF500
	TITLE_ID_JAP = 0x00050000101C5800
	LATEST_VERSION = 64
	
	GAME_SERVER_ID = 0x101C5800
	ACCESS_KEY = "6ef3adf1"
	NEX_VERSION = 31000
	
class Puddle:
	TITLE_ID_EUR = 0x000500001010FB00
	TITLE_ID_USA = 0x0005000010110500
	LATEST_VERSION = 16
	
	GAME_SERVER_ID = 0x1010FB00
	ACCESS_KEY = "afcffb5c"
	NEX_VERSION = 30001 #3.0.1.4

class SMM:
	TITLE_ID_EUR = 0x000500001018DD00
	TITLE_ID_USA = 0x000500001018DC00
	TITLE_ID_JAP = 0x000500001018DB00
	LATEST_VERSION = 272
	
	GAME_SERVER_ID = 0x1018DB00
	ACCESS_KEY = "9f2b4678"
	NEX_VERSION = 30810 #3.8.10 (AMA patch)
	

#===== Switch Games =====#

class ACNH:
	TITLE_ID = 0x01006F8002326000
	LATEST_VERSION = 0x1A0000
	
	GAME_SERVER_ID = 0x2EE2E300
	ACCESS_KEY = "v43a10em"
	NEX_VERSION = 40604 #4.6.4
	CLIENT_VERSION = 2
	
class GBG:
	TITLE_ID = 0x0100FA5010788000
	LATEST_VERSION = 0x10000
	
	GAME_SERVER_ID = 0x2E99DD00
	ACCESS_KEY = "97b08aad"
	NEX_VERSION = 40610
	CLIENT_VERSION = 2

class MK8Deluxe:
	TITLE_ID = 0x0100152000022000
	LATEST_VERSION = 0xA0000
	
	GAME_SERVER_ID = 0x2B309E01
	ACCESS_KEY = "09c1c475"
	NEX_VERSION = 40302

class SMM2:
	TITLE_ID = 0x01009B90006DC000
	LATEST_VERSION = 0x50000
	
	GAME_SERVER_ID = 0x22306D00
	ACCESS_KEY = "fdf6617f"
	NEX_VERSION = 40605 #4.6.25 (appslop)
	CLIENT_VERSION = 60
	
	PIA_VERSION = 51898
	PIA_APP_VERSION = 3
	PIA_KEY = bytes.fromhex("667c18475889faab61f93ef1da180971")

class Splatoon2:
	TITLE_ID = 0x0100F8F0000A2000
	LATEST_VERSION = 0x310000
	
	GAME_SERVER_ID = 0x2DF33D01
	ACCESS_KEY = "4eb18d39"
	NEX_VERSION = 40311
	
	PIA_VERSION = 50903
	PIA_KEY = bytes.fromhex("ee182a63e216cdb1f51ad4bed8cf6508")

class Tetris99:
	TITLE_ID = 0x010040600C5CE000
	LATEST_VERSION = 0x60000
	
	GAME_SERVER_ID = 0x23BDA200
	ACCESS_KEY = "cdd6114d"
	NEX_VERSION = 40600 #4.6.4 (app99)
	CLIENT_VERSION = 5
